# Plot of chaotic|stable area proportions for a 2D 
# optical lattice hamiltonian system:
#
#   H = px**2 + py**2 + V(x,y)
#   V = U*(cos^2(x) + cos^2(y) + 2 a cos(x)cos(y))
#
#   Reads the [area.dat] file with area points 
#   (E, a_s, a_c, a_t, a_s_norm, a_c_norm) obtained 
#   from the SALI method 
#
#   Return the plot of area per energy
#
#   python3.6 plot_area.py
#
#   Implemented by M. Lazarotto (26/05/2020)

import os
import argparse
import numpy as np
from matplotlib import cm
import matplotlib.pyplot as plt
from fractions import Fraction as frac
from matplotlib.ticker import FuncFormatter
from matplotlib.ticker import MultipleLocator
from matplotlib.ticker import FormatStrFormatter

parser = argparse.ArgumentParser()
# Arg: connect points with lines (lines | points)
parser.add_argument('-fl', help = 'File name ',
                    action = 'store', default = '', dest = 'flname')

args = parser.parse_args()

# Potential function
def V(x, y, U, a):
    return U*(np.cos(x)**2 + np.cos(y)**2 + 2 * a * np.cos(x)*np.cos(y))

# Data parse function
def data_parse(flname):
    with open(flname, 'r') as inpt:
        
        dat = []

        for line in inpt:        
            if line[:5] == '# [E]':
                break
            elif ' Emin =' in line:
                Emin = float(line.split()[-1])
            elif ' Emax =' in line:
                Emax = float(line.split()[-1])
            elif ' alpha =' in line:
                a = float(line.split()[-1])
            elif ' nFrm =' in line:
                nFrm = int(line.split()[-1])
                    
        E, A_t = [], []
        A_s, A_c = [], []
        A_s_norm, A_c_norm = [], []
      
        while (line[:3] == '# [' and line != ''):
            
            line = inpt.readline()
            
            while (line[:3] != '# [' and line != ''):
                E.append(float(line.split()[0]))
                A_s.append(float(line.split()[1]))
                A_c.append(float(line.split()[2]))
                A_t.append(float(line.split()[3]))
                A_s_norm.append(float(line.split()[4]))
                A_c_norm.append(float(line.split()[5]))  
                
                line = inpt.readline()

        dat.append(E)
        dat.append(A_s)
        dat.append(A_c)
        dat.append(A_t)
        dat.append(A_s_norm)
        dat.append(A_c_norm)
        dat.append({'Emin': Emin, 'Emax': Emax, 'a': a, 'nFrm': nFrm})

        return dat

### Plot function ###
def plot_routine(dat, norm=True, a=1.0):
    plt.rc('text', usetex = True)  # LaTeX text
    plt.rc('font', family = 'serif', size = 60)
    fig, ax = plt.subplots(figsize = (30, 28))

    Xmin = min(dat[0])
    Xmax = max(dat[0])
    if norm:
        Ymin = 0.0
        Ymax = 1.05
        YMt = MultipleLocator(0.20)            # Y major ticks
        Ymt = MultipleLocator(0.05)            # Y minor ticks
        Y_format = FormatStrFormatter('%0.1f') # decimal labels
    else:
        Ymin = 0.0
        Ymax = 1.1*max(dat[3])
        YMt = MultipleLocator(abs(Ymax - Ymin) / 5.0)   # X major ticks
        Ymt = MultipleLocator(abs(Ymax - Ymin) / 20.0)  # X minor ticks
        Y_format = FormatStrFormatter('%0.1f')          # decimal labels

    # Set X axes #
    XMt = MultipleLocator(abs(Xmax - Xmin) / 8.00)  # X major ticks
    Xmt = MultipleLocator(abs(Xmax - Xmin) / 16.00)  # X minor ticks 
    X_format = FormatStrFormatter('%0.0f')          # decimal labels
    ax.set_xlim([Xmin, Xmax])
    ax.xaxis.set_major_locator(XMt)
    ax.xaxis.set_minor_locator(Xmt)
    ax.xaxis.set_major_formatter(X_format)
    ax.set_xlabel(r'$E$', rotation = 0, labelpad = 15, fontsize = 82)

    # Set Y axes #
    ax.set_ylim([Ymin, Ymax])
    ax.yaxis.set_major_locator(YMt)
    ax.yaxis.set_minor_locator(Ymt)
    ax.yaxis.set_major_formatter(Y_format)
    ax.set_ylabel(r'$A$', rotation = 90, labelpad = 50, fontsize = 72)

    # Set global features #
    plt.tick_params(which = 'major', direction = 'in', length = 18, width = 2.8, 
                    bottom = True, top = True, left = True, right = True, pad = 20)
    plt.tick_params(which = 'minor', direction = 'in', length = 15, width = 2.4, 
                    bottom = True, top = True, left = True, right = True)
    for axis in ['top','bottom','left','right']:
        ax.spines[axis].set_linewidth(4.5)
    
    # Plot curve
    color_sum = 'darkgoldenrod'
    color_chaos = 'black'
    color_stab = 'gold'
    if (norm == True):
        plt.plot(dat[0], dat[4], linestyle = '-', linewidth = 5.0, color = color_stab, zorder = 3)
        plt.plot(dat[0], dat[5], linestyle = '-', linewidth = 5.0, color = color_chaos, zorder = 2)
        plt.plot([Xmin, Xmax], [1.0, 1.0], linestyle = '--', linewidth = 5.0, color = color_sum, zorder = 1, label = r'Total')
        plt.scatter(dat[0], dat[4], marker = 'o', s = 400, c = color_stab, label = r'Regular', zorder = 4)
        plt.scatter(dat[0], dat[5], marker = 'o', s = 400, c = color_chaos, label = r'Chaotic', zorder = 5)

        plt.title(r'Absolute Chaotic/Stable area ($\alpha =$%.3f)' % a, 
                  x = 0.5, y = 1.03, fontsize = 50)
        plt.legend(loc = 'best')
        plt.savefig('area_norm.png')
    else:
        plt.plot(dat[0], dat[1], linestyle = '-', linewidth = 5.0, color = color_stab , zorder = 3)
        plt.plot(dat[0], dat[2], linestyle = '-', linewidth = 5.0, color = color_chaos, zorder = 2)
        plt.plot(dat[0], dat[3], linestyle = '--', linewidth = 5.0, color = color_sum, zorder = 1)
        plt.scatter(dat[0], dat[1], marker = 'o', s = 400, c = color_stab, label = r'Regular', zorder = 6)
        plt.scatter(dat[0], dat[2], marker = 'o', s = 400, c = color_chaos, label = r'Chaotic', zorder = 5)
        plt.scatter(dat[0], dat[3], marker = 'o', s = 400, c = color_sum, label = r'Total', zorder = 4)
        
        plt.title(r'Absolute Chaotic/Stable area ($\alpha =$%.3f)' % a, 
                  x = 0.5, y = 1.03, fontsize = 50)
        plt.legend(loc = 'best')
        plt.savefig('area_abs.png')
    
    plt.clf()
    plt.close()


### Main ###
wd = os.getcwd()
dat = data_parse(wd + '/' + args.flname)
plot_routine(dat, norm=True, a = dat[-1]['a'])
plot_routine(dat, norm=False, a = dat[-1]['a'])