# Plot of SALI for PSS of the 2D optical lattice 
# hamiltonian system:
#
#   H = px**2 + py**2 + V(x,y)
#   V = U (cos**2(x) + cos**2(y) + 2 a cos(x) cos(y))
#
#   Read .dat files with SALI index points 
#   along with its position in the PSS.
#
#   Return the heat map plot (x, px)
#
#   python3.6 plot_sali.py
#
#   Matheus J. Lazarotto (22/05/2020)

import os
import numpy as np
from matplotlib import cm
import matplotlib.pyplot as plt
from fractions import Fraction as frac
from matplotlib.colors import LogNorm
from matplotlib.ticker import LogLocator
from matplotlib.ticker import FuncFormatter
from matplotlib.ticker import MultipleLocator
from matplotlib.ticker import FormatStrFormatter

# Aux. function to use strings of pi
#       format (\frac{num}{den}\pi)
def lbl_pi(val, pos):
    N = np.round(val/np.pi, 3)
    num = int(frac(N).limit_denominator(50).numerator)
    den = int(frac(N).limit_denominator(50).denominator)

    if (num == 0):
        return r'$0$'
    elif (num == den):
        return r'$\pi$'
    elif (num == -den):
        return r'$-\pi$'
    elif (den == 1):
        return r'$%d\pi$' % num
    elif (num == 1):
        return r'$\frac{\pi}{%d}$' % den
    elif (num == -1):
        return r'$-\frac{\pi}{%d}$' % den
    else:
        return r'$\frac{%d}{%d}\pi$' % (num, den)

# Return the order of magnitude of a number
def mag_order(val):
    return np.floor(np.log10(val))

# Potential function
def V(x, y, U, a):
    return U*(np.cos(x)**2 + np.cos(y)**2 + 2 * a * np.cos(x)*np.cos(y))

# PSS file data parse function
def data_parse(flname, compact = False, scl = 'log', 
               scl_min = 1e-16, scl_max = np.sqrt(2), compN = 4):
    with open(flname, 'r') as inpt:
        
        dat = []
        
        for line in inpt:        
            if ' E =' in line:
                E = float(line.split()[-1])
            elif ' U =' in line:
                U = float(line.split()[-1])
            elif ' a =' in line:
                a = float(line.split()[-1])
            elif ' cut_t =' in line:
                cut_t = float(line.split()[-1])
            elif ' displ =' in line:
                displ = float(line.split()[-1])
            elif ' sali_c =' in line:
                sali_c = float(line.split()[-1])
            elif ' mi_s =' in line:
                mi_s = float(line.split()[-1])
            elif ' Ni =' in line:
                Ni = int(line.split()[-1])
            elif ' Nj =' in line:
                Nj = int(line.split()[-1])
            elif '# [SALI]' in line:
                break

        # Initialize T map array
        datS = [[np.nan for i in range(Ni+1)] for j in range(Nj+1)] #----- testar mudar para Ni, Nj (ver o j=Nj-....
        datX, datPx = [], []
        min_s, max_s = 1.0e12, -1.0
        
        order_min = mag_order(scl_min)
        order_max = mag_order(scl_max)
        lims_lin = np.linspace(scl_min, scl_max, compN)
        lims_log = np.logspace(order_min, order_max, compN)

        while (line[:3] == '# [' and line != ''):
            line = inpt.readline()
            while (line[:3] != '# [' and line != ''):
                i = int(line.split()[1])
                j = Nj - int(line.split()[2])
                val = float(line.split()[0])
                
                if (val >= 0.0):
                    if (compact == True):
                        if (scl == 'log'):
                            for k in range(len(lims_log) - 1):
                                up_lim = lims_log[k+1]
                                bottom_lim = lims_log[k]
                                if (bottom_lim < val < up_lim):
                                    datS[j][i] = up_lim
                                elif (lims_log[-1] < val):
                                    datS[j][i] = lims_log[-1]
                                elif (lims_log[0] > val):
                                    datS[j][i] = lims_log[0]

                        elif (scl == 'linear'):
                            for k in range(len(lims_lin)-1):
                                up_lim = lims_lin[k+1]
                                bottom_lim = lims_lin[k]
                                if (bottom_lim < val < up_lim):
                                    datS[j][i] = (up_lim + bottom_lim) / 2
                    else:
                        if (val == 0.0):
                            datS[j][i] = scl_min
                        else:
                            datS[j][i] = val
                    
                    if (val < min_s):
                        min_s = val
                    if (val > max_s):
                        max_s = val
                else:
                    datS[j][i] = np.nan
                
                datX.append(float(line.split()[4]))
                datPx.append(float(line.split()[6]) / np.sqrt(E))

                line = inpt.readline()

        dat.append(datS)
        dat.append(datX)
        dat.append(datPx)
        dat.append({'U': U, 'a': a, 'E': E, 'mi_s': mi_s, 'cut_t': cut_t, 
                    'sali_c': sali_c, 'displ': displ, 'Ni': Ni, 'Nj': Nj, 
                    'min_s': min_s, 'max_s': max_s})

        return dat

### Plot routine ###
def print_routine(dat, scl_min, scl_max, 
                  Xmin=0.0, Xmax=np.pi, 
                  Ymin=-1.0, Ymax=1.0):
    # Executes the plot
    plt.rc('text', usetex = True)  # LaTeX text
    plt.rc('font', family = 'serif', size = 80)
    fig, ax = plt.subplots(figsize = (38, 28))

    U = dat[-1]['U']
    a = dat[-1]['a']
    E = dat[-1]['E']
    Ni = dat[-1]['Ni']
    Nj = dat[-1]['Nj']
    mi_s = dat[-1]['mi_s']
    displ = dat[-1]['displ']
    cut_t = dat[-1]['cut_t']
    min_s = dat[-1]['min_s']
    max_s = dat[-1]['max_s']
    sali_c = dat[-1]['sali_c']

    # Set X axes #
    XMt = MultipleLocator(abs(Xmax - Xmin) / 4.0)  # X major ticks
    Xmt = MultipleLocator(abs(Xmax - Xmin) / 12.0) # X minor ticks
    X_format = FuncFormatter(lbl_pi)               # string labels
    ax.set_xlim([Xmin, Xmax])
    ax.xaxis.set_major_locator(XMt)
    ax.xaxis.set_minor_locator(Xmt)
    ax.xaxis.set_major_formatter(X_format)
    ax.set_xlabel(r'$x$', rotation = 0, labelpad = 20, fontsize = 82)

    # Set Px axes #
    YMt = MultipleLocator(abs(Ymax - Ymin) / 4.0)  # Px major ticks
    Ymt = MultipleLocator(abs(Ymax - Ymin) / 12.0) # Px minor ticks 
    Y_format = FormatStrFormatter('%0.1f')         # decimal labels
    ax.set_ylim([Ymin, Ymax])
    ax.yaxis.set_major_locator(YMt)
    ax.yaxis.set_minor_locator(Ymt)
    ax.yaxis.set_major_formatter(Y_format)
    ax.set_ylabel(r'$p_x$', rotation = 90, labelpad = 15, fontsize = 82)

    # Set global features #
    ax.tick_params(which = 'major', direction = 'in', length = 22, width = 3.0, 
                    bottom = True, top = True, left = True, right = True, pad = 20)
    ax.tick_params(which = 'minor', direction = 'in', length = 18, width = 2.4, 
                    bottom = True, top = True, left = True, right = True)
    for axis in ['top','bottom','left','right']:
        ax.spines[axis].set_linewidth(4.5)

    # Create heatmap
    extent = [Xmin, Xmax, Ymin, Ymax]
    plt.imshow(dat[0], extent=extent, cmap = 'gnuplot', interpolation = 'none', aspect = 'auto', origin = 'lower',
                       norm=LogNorm(vmin=scl_min, vmax=scl_max))
    
    cbar = plt.colorbar()
    cbar.ax.tick_params(which = 'major', direction = 'in', length = 20, width = 3.0, pad = 20)
    for axis in ['top','bottom','left','right']:
        cbar.ax.spines[axis].set_linewidth(4.5)

    plt.text(0.75, 0.93, r'$\alpha =%.1f$' % a,
            transform = ax.transAxes, fontsize = 100)
    plt.text(0.02, 0.93, r'$E =%.1f$' % E,
            transform = ax.transAxes, fontsize = 100)

    plt.title('SALI (Smaller Alignment Index) in PSS ' + 
            r'$(y = $' + '%s' % lbl_pi(mi_s, 0) + r'$; \quad p_y>0)$'  + '\n' + 
            r'$U =$%.0f' % U + r'$\quad\quad E =$%.5f' % E + r'$\quad\quad \alpha =$%.5f' % a + '\n' +
            r'$d_0 =$%.1e' % displ + r'$\quad\quad s_{\textrm{cut}} =$%.1e' % sali_c + r'$\quad\quad t_{\textrm{max}} =$%.1f' % cut_t + 
            r'$\quad\quad N_x =$%d' % Ni + r'$\quad\quad N_{px} =$%d' % Nj,
            y=1.02, fontsize = 40)

    plt.savefig('sali_a' + '{:.4f}'.format(a) + '_e' + '{:.4f}'.format(E) + '.png',
                bbox_inches='tight')
    plt.close(fig)
    plt.clf()


## Main ##
compact = False
scl = 'log'
scl_min = 1e-14
scl_max = np.sqrt(2)
compact_N = 5

wd = os.getcwd()
for fl in os.listdir(wd):
    if fl.endswith('.dat') and 'sali_a' in fl:
        dat = data_parse(fl, compact, scl, scl_min, scl_max, compact_N)
        print_routine(dat, scl_min, scl_max)
