#include "ode_sys_solv.hpp"
#include "auxf.hpp"

void print_Vf(vector<double> v) {
    cout << "[";
    for (unsigned int i = 0; i < v.size(); i++) { 
        if (i == (v.size() - 1)) { cout << v[i]; }
        else { cout << v[i] << ", "; }
    }
    cout << "]\n" << endl;
}

vector<double> arr_to_vec(double arr[], unsigned int n)
{
    vector<double> vec(n);
    
    for (unsigned int k = 0; k < n; k++) { vec[k] = arr[k]; }

    return vec;
}

vector<int> linspace_i(int xi, int xf, unsigned int skip) 
{
    vector<int> v;
    assert(xi <= xf);

    if (abs(xf - xi) < (int) skip) { v.push_back(xi); }
    else
    {
        for (int n = 0; n < abs(xf - xi) / (int) skip; n++)
        {
            v.push_back(xi + n * skip);
        }
    }

    return v;
}

vector<double> linspace_f(double xi, double xf, unsigned int N) 
{
    vector<double> v(N+1);

    for (int n = 0; n <= (int) N; n++)
    {
        v[n] = xi + ((double) n / (double) N) * (xf - xi);
    }

    return v;
}

double norm(vector<double> v)
{
    double n = 0.0;
    
    for (auto vi: v) { n += vi * vi; }
    
    return sqrt(n);
}

void normalize(vector<double> &v)
{
    /* Normalize a vector */
    double n = norm(v);
    
    assert(n != 0.0);

    for (unsigned int k = 0; k < v.size(); k++) { v[k] = v[k] / n; }
}

vector<double> lincomb_f(double a, vector<double> v1,
                         double b, vector<double> v2)
{
    vector<double> lin(v1.size());

    assert(v1.size() == v2.size());

    for (unsigned int k = 0; k < v1.size(); k++) 
    { 
        lin[k] = a * v1[k] + b * v2[k]; 
    }

    return lin;
}

double V(double x, double y, double U, double a) 
{
    return U * (pow(cos(x), 2) + pow(cos(y), 2) + 2 * a * cos(x) * cos(y));
}

double Ham(double x, double y, double px, double py, double U, double a) 
{
    return pow(px, 2) + pow(py, 2) + V(x, y, U, a);
}

double dqdt(double pi) 
{ 
    return 2 * pi; 
}

double dpdt(double qi, double qj, double U, double a) 
{ 
    return 2 * U * sin(qi) * (cos(qi) + a * cos(qj)); 
}

double Vi(double qi, double qj, double U, double a) 
{
    return -2.0 * U * sin(qi) * (cos(qi) + a * cos(qj));
}

double Vii(double qi, double qj, double U, double a) 
{
    return -2.0 * U * (cos(2*qi) + a * cos(qi) * cos(qj));
}

double Vij(double qi, double qj, double U, double a) 
{
    return 2.0 * U * a *sin(qi) * sin(qj); 
}

double Viii(double qi, double qj, double U, double a)
{
    return 2.0 * U * (2.0 * sin(2.0 * qi) + a * sin(qi) * cos(qj));
}

double Viij(double qi, double qj, double U, double a)
{
    return 2.0 * U * a * cos(qi) * sin(qj);
}

double Viji(double qi, double qj, double U, double a)
{
    return 2.0 * U * a * cos(qi) * sin(qj);
}

int system_funct(double t, const double s[], double f[], void *args) 
{
    (void)(t); // Avoid unused parameter warning
    double *par = (double *)args;
    double a = par[0];
    double U = par[1];
    
    f[0] = dqdt(s[2]);
    f[1] = dqdt(s[3]);
    f[2] = dpdt(s[0], s[1], U, a);
    f[3] = dpdt(s[1], s[0], U, a);

    return GSL_SUCCESS; 
}

int system_jacob(double t, const double s[], double *dfdy, double dfdt[], void *args) 
{
    (void)(t); // Avoid unused parameter warning
    double *par = (double *)args;
    double a = par[0];
    double U = par[1];

    gsl_matrix_view dfdy_mat = gsl_matrix_view_array(dfdy, 4, 4);
    gsl_matrix *m = &dfdy_mat.matrix;

    gsl_matrix_set(m, 0, 0, 0.0);
    gsl_matrix_set(m, 0, 1, 0.0);
    gsl_matrix_set(m, 0, 2, 2.0);
    gsl_matrix_set(m, 0, 3, 0.0);
    
    gsl_matrix_set(m, 1, 0, 0.0);
    gsl_matrix_set(m, 1, 1, 0.0);
    gsl_matrix_set(m, 1, 2, 0.0);
    gsl_matrix_set(m, 1, 3, 2.0);
    
    gsl_matrix_set(m, 2, 0, -Vii(s[0], s[1], U, a));
    gsl_matrix_set(m, 2, 1, -Vij(s[0], s[1], U, a));
    gsl_matrix_set(m, 2, 2, 0.0);
    gsl_matrix_set(m, 2, 3, 0.0);
    
    gsl_matrix_set(m, 3, 0, -Vij(s[1], s[0], U, a));
    gsl_matrix_set(m, 3, 1, -Vii(s[1], s[0], U, a));
    gsl_matrix_set(m, 3, 2, 0.0);
    gsl_matrix_set(m, 3, 3, 0.0);

    dfdt[0] = 0.0; 
    dfdt[1] = 0.0;
    dfdt[2] = 0.0;
    dfdt[3] = 0.0;

    return GSL_SUCCESS;
}

int system_funct_lin(double t, const double s[], double f[], void *args)
{
    (void)(t); // Avoid unused parameter warning
    double *par = (double *)args;
    double a = par[0];
    double U = par[1];
    double x0 = par[2];
    double y0 = par[3];
    
    /** Linearized system used for 
     *  evolving the deviation 
     *  vectors.
     **/
    /* s[0] = v_x
       s[1] = v_y
       s[2] = v_px
       s[3] = v_py */
    f[0] = 2.0 * s[2];
    f[1] = 2.0 * s[3];
    f[2] = -s[0] * Vii(x0, y0, U, a) - s[1] * Vij(x0, y0, U, a);
    f[3] = -s[0] * Vij(x0, y0, U, a) - s[1] * Vii(y0, x0, U, a);

    return GSL_SUCCESS;
}

int system_jacob_lin(double t, const double s[], double *dfdy, double dfdt[], void *args)
{
    (void)(t); // Avoid unused parameter warning
    double *par = (double *)args;
    double a = par[0];
    double U = par[1];
    double x0 = par[2];
    double y0 = par[3];

    gsl_matrix_view dfdy_mat = gsl_matrix_view_array(dfdy, 4, 4);
    gsl_matrix *m = &dfdy_mat.matrix;

    gsl_matrix_set(m, 0, 0, 0.0);
    gsl_matrix_set(m, 0, 1, 0.0);
    gsl_matrix_set(m, 0, 2, 2.0);
    gsl_matrix_set(m, 0, 3, 0.0);
    
    gsl_matrix_set(m, 1, 0, 0.0);
    gsl_matrix_set(m, 1, 1, 0.0);
    gsl_matrix_set(m, 1, 2, 0.0);
    gsl_matrix_set(m, 1, 3, 2.0);
    
    gsl_matrix_set(m, 2, 0, -Vii(x0, y0, U, a));
    gsl_matrix_set(m, 2, 1, -Vij(x0, y0, U, a));
    gsl_matrix_set(m, 2, 2, 0.0);
    gsl_matrix_set(m, 2, 3, 0.0);
    
    gsl_matrix_set(m, 3, 0, -Vij(y0, x0, U, a));
    gsl_matrix_set(m, 3, 1, -Vii(y0, x0, U, a));
    gsl_matrix_set(m, 3, 2, 0.0);
    gsl_matrix_set(m, 3, 3, 0.0);

    dfdt[0] = 0.0; 
    dfdt[1] = 0.0;
    dfdt[2] = 0.0;
    dfdt[3] = 0.0;

    return GSL_SUCCESS;
}

int system_funct_all(double t, const double s[], double f[], void *args)
{
    (void)(t); // Avoid unused parameter warning
    double *par = (double *)args;
    double a = par[0];
    double U = par[1];

    /**
     * s[0] = x
     * s[1] = y
     * s[2] = px
     * s[3] = py
     * 
     * s[4] = v1_x
     * s[5] = v1_y
     * s[6] = v1_px
     * s[7] = v1_py
     * 
     * s[8] = v2_x
     * s[9] = v2_y
     * s[10] = v2_px
     * s[11] = v2_py
     **/

    f[0] = dqdt(s[2]);
    f[1] = dqdt(s[3]);
    f[2] = dpdt(s[0], s[1], U, a);
    f[3] = dpdt(s[1], s[0], U, a);

    f[4] = 2.0 * s[6];
    f[5] = 2.0 * s[7];
    f[6] = -s[4] * Vii(s[0], s[1], U, a) - s[5] * Vij(s[0], s[1], U, a);
    f[7] = -s[4] * Vij(s[0], s[1], U, a) - s[5] * Vii(s[1], s[0], U, a);

    f[8] = 2.0 * s[10];
    f[9] = 2.0 * s[11];
    f[10] = -s[8] * Vii(s[0], s[1], U, a) - s[9] * Vij(s[0], s[1], U, a);
    f[11] = -s[8] * Vij(s[0], s[1], U, a) - s[9] * Vii(s[1], s[0], U, a);

    return GSL_SUCCESS;
}

int system_jacob_all(double t, const double s[], double *dfdy, double dfdt[], void *args)
{
    (void)(t); // Avoid unused parameter warning
    double *par = (double *)args;
    double a = par[0];
    double U = par[1];

    gsl_matrix_view dfdy_mat = gsl_matrix_view_array(dfdy, 12, 12);
    gsl_matrix *m = &dfdy_mat.matrix;

    gsl_matrix_set(m, 0, 0,  0.0);
    gsl_matrix_set(m, 0, 1,  0.0);
    gsl_matrix_set(m, 0, 2,  2.0);
    gsl_matrix_set(m, 0, 3,  0.0);
    gsl_matrix_set(m, 0, 4,  0.0);
    gsl_matrix_set(m, 0, 5,  0.0);
    gsl_matrix_set(m, 0, 6,  0.0);
    gsl_matrix_set(m, 0, 7,  0.0);
    gsl_matrix_set(m, 0, 8,  0.0);
    gsl_matrix_set(m, 0, 9,  0.0);
    gsl_matrix_set(m, 0, 10, 0.0);
    gsl_matrix_set(m, 0, 11, 0.0);

    gsl_matrix_set(m, 1, 0,  0.0);
    gsl_matrix_set(m, 1, 1,  0.0);
    gsl_matrix_set(m, 1, 2,  0.0);
    gsl_matrix_set(m, 1, 3,  2.0);
    gsl_matrix_set(m, 1, 4,  0.0);
    gsl_matrix_set(m, 1, 5,  0.0);
    gsl_matrix_set(m, 1, 6,  0.0);
    gsl_matrix_set(m, 1, 7,  0.0);
    gsl_matrix_set(m, 1, 8,  0.0);
    gsl_matrix_set(m, 1, 9,  0.0);
    gsl_matrix_set(m, 1, 10, 0.0);
    gsl_matrix_set(m, 1, 11, 0.0);

    gsl_matrix_set(m, 2, 0, -Vii(s[0], s[1], U, a));
    gsl_matrix_set(m, 2, 1, -Vij(s[0], s[1], U, a));
    gsl_matrix_set(m, 2, 2,  0.0);
    gsl_matrix_set(m, 2, 3,  0.0);
    gsl_matrix_set(m, 2, 4,  0.0);
    gsl_matrix_set(m, 2, 5,  0.0);
    gsl_matrix_set(m, 2, 6,  0.0);
    gsl_matrix_set(m, 2, 7,  0.0);
    gsl_matrix_set(m, 2, 8,  0.0);
    gsl_matrix_set(m, 2, 9,  0.0);
    gsl_matrix_set(m, 2, 10, 0.0);
    gsl_matrix_set(m, 2, 11, 0.0);

    gsl_matrix_set(m, 3, 0, -Vij(s[1], s[0], U, a));
    gsl_matrix_set(m, 3, 1, -Vii(s[1], s[0], U, a));
    gsl_matrix_set(m, 3, 2,  0.0);
    gsl_matrix_set(m, 3, 3,  0.0);
    gsl_matrix_set(m, 3, 4,  0.0);
    gsl_matrix_set(m, 3, 5,  0.0);
    gsl_matrix_set(m, 3, 6,  0.0);
    gsl_matrix_set(m, 3, 7,  0.0);
    gsl_matrix_set(m, 3, 8,  0.0);
    gsl_matrix_set(m, 3, 9,  0.0);
    gsl_matrix_set(m, 3, 10, 0.0);
    gsl_matrix_set(m, 3, 11, 0.0);

    gsl_matrix_set(m, 4, 0,  0.0);
    gsl_matrix_set(m, 4, 1,  0.0);
    gsl_matrix_set(m, 4, 2,  0.0);
    gsl_matrix_set(m, 4, 3,  0.0);
    gsl_matrix_set(m, 4, 4,  0.0);
    gsl_matrix_set(m, 4, 5,  0.0);
    gsl_matrix_set(m, 4, 6,  2.0);
    gsl_matrix_set(m, 4, 7,  0.0);
    gsl_matrix_set(m, 4, 8,  0.0);
    gsl_matrix_set(m, 4, 9,  0.0);
    gsl_matrix_set(m, 4, 10, 0.0);
    gsl_matrix_set(m, 4, 11, 0.0);

    gsl_matrix_set(m, 5, 0,  0.0);
    gsl_matrix_set(m, 5, 1,  0.0);
    gsl_matrix_set(m, 5, 2,  0.0);
    gsl_matrix_set(m, 5, 3,  0.0);
    gsl_matrix_set(m, 5, 4,  0.0);
    gsl_matrix_set(m, 5, 5,  0.0);
    gsl_matrix_set(m, 5, 6,  0.0);
    gsl_matrix_set(m, 5, 7,  2.0);
    gsl_matrix_set(m, 5, 8,  0.0);
    gsl_matrix_set(m, 5, 9,  0.0);
    gsl_matrix_set(m, 5, 10, 0.0);
    gsl_matrix_set(m, 5, 11, 0.0);

    gsl_matrix_set(m, 6, 0, -s[4] * Viii(s[0], s[1], U, a) - s[5] * Viji(s[0], s[1], U, a));
    gsl_matrix_set(m, 6, 1, -s[4] * Viij(s[0], s[1], U, a) - s[5] * Viji(s[1], s[0], U, a));
    gsl_matrix_set(m, 6, 2,  0.0);
    gsl_matrix_set(m, 6, 3,  0.0);
    gsl_matrix_set(m, 6, 4, -Vii(s[0], s[1], U, a));
    gsl_matrix_set(m, 6, 5, -Vij(s[0], s[1], U, a));
    gsl_matrix_set(m, 6, 6,  0.0);
    gsl_matrix_set(m, 6, 7,  0.0);
    gsl_matrix_set(m, 6, 8,  0.0);
    gsl_matrix_set(m, 6, 9,  0.0);
    gsl_matrix_set(m, 6, 10, 0.0);
    gsl_matrix_set(m, 6, 11, 0.0);

    gsl_matrix_set(m, 7, 0, -s[4] * Viji(s[0], s[1], U, a) - s[5] * Viij(s[1], s[0], U, a));
    gsl_matrix_set(m, 7, 1, -s[4] * Viji(s[1], s[0], U, a) - s[5] * Viii(s[1], s[0], U, a));
    gsl_matrix_set(m, 7, 2,  0.0);
    gsl_matrix_set(m, 7, 3,  0.0);
    gsl_matrix_set(m, 7, 4, -Vij(s[1], s[0], U, a));
    gsl_matrix_set(m, 7, 5, -Vii(s[1], s[0], U, a));
    gsl_matrix_set(m, 7, 6,  0.0);
    gsl_matrix_set(m, 7, 7,  0.0);
    gsl_matrix_set(m, 7, 8,  0.0);
    gsl_matrix_set(m, 7, 9,  0.0);
    gsl_matrix_set(m, 7, 10, 0.0);
    gsl_matrix_set(m, 7, 11, 0.0);

    gsl_matrix_set(m, 8, 0,  0.0);
    gsl_matrix_set(m, 8, 1,  0.0);
    gsl_matrix_set(m, 8, 2,  0.0);
    gsl_matrix_set(m, 8, 3,  0.0);
    gsl_matrix_set(m, 8, 4,  0.0);
    gsl_matrix_set(m, 8, 5,  0.0);
    gsl_matrix_set(m, 8, 6,  0.0);
    gsl_matrix_set(m, 8, 7,  0.0);
    gsl_matrix_set(m, 8, 8,  0.0);
    gsl_matrix_set(m, 8, 9,  0.0);
    gsl_matrix_set(m, 8, 10, 2.0);
    gsl_matrix_set(m, 8, 11, 0.0);

    gsl_matrix_set(m, 9, 0,  0.0);
    gsl_matrix_set(m, 9, 1,  0.0);
    gsl_matrix_set(m, 9, 2,  0.0);
    gsl_matrix_set(m, 9, 3,  0.0);
    gsl_matrix_set(m, 9, 4,  0.0);
    gsl_matrix_set(m, 9, 5,  0.0);
    gsl_matrix_set(m, 9, 6,  0.0);
    gsl_matrix_set(m, 9, 7,  0.0);
    gsl_matrix_set(m, 9, 8,  0.0);
    gsl_matrix_set(m, 9, 9,  0.0);
    gsl_matrix_set(m, 9, 10, 0.0);
    gsl_matrix_set(m, 9, 11, 2.0);

    gsl_matrix_set(m, 10, 0, -s[8] * Viii(s[0], s[1], U, a) - s[9] * Viji(s[0], s[1], U, a));
    gsl_matrix_set(m, 10, 1, -s[8] * Viij(s[0], s[1], U, a) - s[9] * Viji(s[1], s[0], U, a));
    gsl_matrix_set(m, 10, 2,  0.0);
    gsl_matrix_set(m, 10, 3,  0.0);
    gsl_matrix_set(m, 10, 4,  0.0);
    gsl_matrix_set(m, 10, 5,  0.0);
    gsl_matrix_set(m, 10, 6,  0.0);
    gsl_matrix_set(m, 10, 7,  0.0);
    gsl_matrix_set(m, 10, 8, -Vii(s[0], s[1], U, a));
    gsl_matrix_set(m, 10, 9, -Vij(s[0], s[1], U, a));
    gsl_matrix_set(m, 10, 10, 0.0);
    gsl_matrix_set(m, 10, 11, 0.0);

    gsl_matrix_set(m, 11, 0, -s[8] * Viji(s[0], s[1], U, a) - s[9] * Viij(s[1], s[0], U, a));
    gsl_matrix_set(m, 11, 1, -s[8] * Viji(s[1], s[0], U, a) - s[9] * Viii(s[1], s[0], U, a));
    gsl_matrix_set(m, 11, 2,  0.0);
    gsl_matrix_set(m, 11, 3,  0.0);
    gsl_matrix_set(m, 11, 4,  0.0);
    gsl_matrix_set(m, 11, 5,  0.0);
    gsl_matrix_set(m, 11, 6,  0.0);
    gsl_matrix_set(m, 11, 7,  0.0);
    gsl_matrix_set(m, 11, 8, -Vij(s[1], s[0], U, a));
    gsl_matrix_set(m, 11, 9, -Vii(s[1], s[0], U, a));
    gsl_matrix_set(m, 11, 10, 0.0);
    gsl_matrix_set(m, 11, 11, 0.0);

    dfdt[0] = 0.0; 
    dfdt[1] = 0.0;
    dfdt[2] = 0.0;
    dfdt[3] = 0.0;
    
    dfdt[4] = 0.0; 
    dfdt[5] = 0.0;
    dfdt[6] = 0.0;
    dfdt[7] = 0.0;
    
    dfdt[8] = 0.0; 
    dfdt[9] = 0.0;
    dfdt[10] = 0.0;
    dfdt[11] = 0.0;
    
    return GSL_SUCCESS;
}

int pss_henon_step(unsigned int dimS, double mi_s, double varp[], double *args)
{
    double den = dqdt(varp[3]);
    double dmi = mi_s - varp[1];
    double var_pss[dimS];

    var_pss[0] = varp[0] + dmi * dqdt(varp[2]) / den;
    var_pss[1] = mi_s;
    var_pss[2] = varp[2] + dmi * dpdt(varp[0], varp[1], args[1], args[0]) / den;
    var_pss[3] = varp[3] + dmi * dpdt(varp[1], varp[0], args[1], args[0]) / den;

    for (unsigned int k = 0; k < dimS; k++) { varp[k] = var_pss[k]; }
    
    return 0;
}