#include "grid.hpp"

#include <iostream>

Grid::Grid(unsigned int width, unsigned int height,
           double x_min, double x_max, 
           double y_min, double y_max,
           unsigned int pt_dim)
{
    /** This constructor sets the grid sites as a matrix with 
     *  dimensions [width+1]x[height+1]x[pt_dim] within the real 
     *  space (float) boundaries [x_min,x_max]x[y_min,y_max].
     **/

    assert(width > 0);
    assert(height > 0);
    assert(x_max > x_min);
    assert(y_max > y_min);
    assert(pt_dim >= 2);
    
    gWidth = width + 1;
    gHeight = height + 1;
    gPoint_dim = pt_dim;

    sX_min = x_min;
    sX_max = x_max;
    sY_min = y_min;
    sY_max = y_max;

    /** Allocate and Set grid sites.
     *  Points initially setted to a
     *  symbolic 'empty_val = -9999'.
     **/
    grid.reserve(gWidth);
    for (unsigned int i = 0; i < gWidth; i++)
    {
        grid[i].reserve(gHeight);
        for (unsigned int j = 0; j < gHeight; j++)
        {
            grid[i][j].reserve(pt_dim);             
            for (unsigned int p = 0; p < pt_dim; p++)
            {
                grid[i][j][p] = empty_val;
            }
        }
    }
}

bool Grid::check_empty_flt(double sX, double sY)
{
    unsigned int i = (int) ((gWidth-1) * (sX - sX_min) / fabs(sX_max - sX_min));
    unsigned int j = (int) ((gHeight-1) * (1.0 - ((sY - sY_min) / fabs(sY_max - sY_min))));

    assert(i < gWidth);
    assert(j < gHeight);

    bool empty = (grid[i][j][0] == empty_val) ? true : false;

    return empty; 
}

bool Grid::check_empty_ndx(unsigned int i, unsigned int j)
{
    assert(i < gWidth);
    assert(j < gHeight);

    bool empty = (grid[i][(gHeight - 1) - j][0] == empty_val) ? true : false;

    return empty; 
}

bool Grid::check_empty_flt_neigh(double sX, double sY, unsigned int Neig)
{
    unsigned int i = (int) ((gWidth-1) * (sX - sX_min) / fabs(sX_max - sX_min));
    unsigned int j = (int) ((gHeight-1) * (1.0 - ((sY - sY_min) / fabs(sY_max - sY_min))));

    assert(i < gWidth);
    assert(j < gHeight);

    bool empty = true;
    for (int ni = -Neig; ni <= (int) Neig; ni++)
    {
        for (int nj = -Neig; nj <= (int) Neig; nj++)
        {
            if (0 <= (i + ni) && (i + ni) < gWidth &&
                0 <= (j + nj) && (j + nj) < gHeight)
            {
                if (grid[i + ni][j + nj][0] != empty_val)
                {
                    empty = false;
                    return empty;
                }
            }                
        }
    }

    return empty;
}

bool Grid::check_empty_ndx_neigh(unsigned int i, unsigned int j, unsigned int Neig)
{
    assert(i < gWidth);
    assert(j < gHeight);

    bool empty = true;
    for (int ni = -Neig; ni <= (int) Neig; ni++)
    {
        for (int nj = -Neig; nj <= (int) Neig; nj++)
        {
            if (0 <= (i + ni) && (i + ni) < gWidth &&
                0 <= (j + nj) && (j + nj) < gHeight)
            {
                if (grid[i + ni][gHeight - (j + nj)][0] != empty_val)
                {
                    empty = false;
                    return empty;
                }
            }                
        }
    }

    return empty;
}

void Grid::set_grid_pt_flt(double sX, double sY, vector<double> pt)
{
    /** Receives a float position to clip to the grid and
     *  the full coordinate point to attach to it
     **/
    unsigned int i = (int) ((gWidth-1) * (sX - sX_min) / fabs(sX_max - sX_min));
    unsigned int j = (int) ((gHeight-1) * (1.0 - ((sY - sY_min) / fabs(sY_max - sY_min))));

    assert(i < gWidth);
    assert(j < gHeight);
    assert(pt.size() == gPoint_dim);

    for (unsigned int k = 0; k < pt.size(); k++)
    {
        grid[i][j][k] = pt[k];
    }
}

void Grid::set_grid_pt_ndx(unsigned int i, unsigned int j, vector<double> pt)
{
    assert(i < gWidth);
    assert(j < gHeight);
    assert(pt.size() == gPoint_dim);

    for (unsigned int k = 0; k < pt.size(); k++) 
    { 
        grid[i][(gHeight - 1) - j][k] = pt[k];
    }
}

vector<double> Grid::get_grid_pt(unsigned int i, unsigned int j)
{
    assert(i < gWidth);
    assert(j < gHeight);

    vector<double> v(gPoint_dim);
    
    for (unsigned int k = 0; k < gPoint_dim; k++) 
    { 
        v[k] = grid[i][(gHeight-1) - j][k]; 
    }
    
    return v;
}

vector<int> Grid::get_grid_ndx(double sX, double sY)
{
    /** Receives a position in phase space (sX, sY) 
     *  and returns the index position (clipped) on 
     *  the grid. 
     **/
    
    vector<int> grid_pos(2);
    
    unsigned int i = (int) ((gWidth-1) * (sX - sX_min) / fabs(sX_max - sX_min));
    unsigned int j = (int) ((gHeight-1) * (1.0 - ((sY - sY_min) / fabs(sY_max - sY_min))));

    assert(i < gWidth);    
    assert(j < gHeight);

    grid_pos[0] = i;
    grid_pos[1] = j;    

    return grid_pos;
}
