/** SALI method
 *  
 *  The Smaller Alignment Index (SALI) is a numerical method 
 *  developed to quickly distinguish between stable and chaotic 
 *  orbits. For a given initial point, two small displacements 
 *  are made from it and evolved in time, its direction deviance 
 *  is measured to determine between stable or chaotic, specifi-
 *  cally, for a stable orbit the alignment index keeps approxi-
 *  mately constant and positive, whereas for chaotic orbits it 
 *  goes to zero.
 *  The small deviations are evolved according to the linearized 
 *  equation of motion (in tangent space) whereas the reference 
 *  orbit is evolved normally. The deviations are normalized each 
 *  step to avoid overflow due to exponential divergence.
 * 
 *  Reference: [Ch Skokos et al 2004 J. Phys. A: Math. Gen.37 6269]
 * 
 *  This program is an application of the method; It applies the 
 *  index measure for a set of initial points in a Poincaré Sur-
 *  face Section (PSS) for the 2D optical lattice hamiltonian 
 *  system and returns as output the SALI index after a time 
 *  'max_t', as a heat map on the PSS, depicting island and chao-
 *  tic regions.
 * 
 *  To save calculation time, the points are saved into a grid 
 *  object and points that already crossed the PSS are not used
 *  as initial points.
 *  
 *  GSL integration scheme (auxf.cpp | ode_sys_solv.cpp) (fixed time step)
 *    
 *  Usage: The whole program runs directly from 'bash bash_run.sh'
 *         or manually alone by compile with 'make' --> './run_sali'
 *         When running manually, the arguments are given as:
 * 
 *         ./run_sali [iN] [ie] [a] [Emin] [Emax] [Ngrid_x] [Ngrid_px] [nThreads]
 *    
 *  Implemented by: M. Lazarotto (26/05/2020)
 **/

#include <cstdlib>
#include <chrono>
#include <iomanip>
#include <omp.h>

#include "auxf.hpp"
#include "sali.hpp"
#include "grid.hpp"
#include "ode_sys_solv.hpp"

#define PI M_PI

void print_prog_bar(int bar_width, double progress, int ASCII_id);

int main(int argc, char **argv) 
{
    auto start = chrono::steady_clock::now(); /* Set clock */
    
    /** Parse external args **/
    const int iN = stoi(argv[1]);       /* Total energy values   */
    const int ie = stoi(argv[2]);       /* Current energy value  */
    const double a = stof(argv[3]);     /* Coupling parameter    */
    const double Emin = stof(argv[4]);  /* Min energy interval   */
    const double Emax = stof(argv[5]);  /* Max energy interval   */
    const double E = Emin + ((double) ie / (double) iN) * fabs(Emax - Emin);
    const int Ngrid_x = stoi(argv[6]);  /* Subdivisions along x  */
    const int Ngrid_px = stoi(argv[7]); /* Subdivisions along px */
    const int nThreads = stoi(argv[8]); /* # of Threads to run   */

    /** System parameters **/
    const unsigned int dimS = 4;        /* Number of variables (var.)  */
    const double U = 20.0;              /* Intensity                   */
    const double mi_s = PI / 2.0;       /* PSS cross (y = mis; py > 0) */
    const double max_t = 2e3;           /* Max evolution time to convergence */
    const double displ = 1e-8;          /* Displacement around point   */
    const double sali_c = 1e-12;        /* Convergence threshold of sali index */

    /** Set initial conditions in PSS **/
    /** Box limits to gridify **/ 
    const double x_min = 0.0;           /* Min x  of grid      */
    const double x_max =  PI;           /* Max x  of grid      */
    const double px_min = -sqrt(E);     /* Min px of grid      */
    const double px_max =  sqrt(E);     /* Max px of grid      */
    const double mod_inf = -PI;         /* Inferior cell limit */
    const double mod_top =  PI;         /* Superior cell limit */

    /** Set grid **/
    Grid grid(Ngrid_x, Ngrid_px, x_min, x_max, px_min, px_max, dimS);

    /** Measured areas **/
    double a_t = 0.0;                   /* Total area   */
    double a_s = 0.0;                   /* Stable area  */
    double a_c = 0.0;                   /* Chaotic area */
    const double cell_a = fabs(x_max - x_min) * fabs(px_max - px_min) / ((double) Ngrid_x * Ngrid_px);
                                        /* Area per cell */

    /** Out file header **/
    char fname_pts[100];                /* Out file name */
    char fname_ipt[100];                /* Out init. points file name */  
    FILE *fout;                         /* Out file: SALI points */
    FILE *fipt;                         /* Out file: init. points */
    sprintf(fname_pts, "sali_a%.3f_e%.3f.dat", a, E);
    sprintf(fname_ipt, "sali_ipt_a%.3f_e%.3f.dat", a, E);
    fout = fopen(fname_pts, "w");
    fipt = fopen(fname_ipt, "w");
    fprintf(fout, "# SALI times in PSS (y = mi_s;  py>0) \n");
    fprintf(fipt, "# SALI times in PSS (y = mi_s;  py>0) \n");
    fprintf(fout, "# H(x,y) = px**2 + py**2 + U(cos**2(x) + cos**2(y) + 2*a*cos(x)*cos(y))\n");
    fprintf(fipt, "# H(x,y) = px**2 + py**2 + U(cos**2(x) + cos**2(y) + 2*a*cos(x)*cos(y))\n");
    fprintf(fout, "# U = %f\n", U); 
    fprintf(fipt, "# U = %f\n", U); 
    fprintf(fout, "# a = %f\n", a);
    fprintf(fipt, "# a = %f\n", a);
    fprintf(fout, "# E = %.4f * Vmax(U,a) = %f\n", E / V(0.0,0.0,U,a), E);
    fprintf(fipt, "# E = %.4f * Vmax(U,a) = %f\n", E / V(0.0,0.0,U,a), E);
    fprintf(fout, "# mi_s = %f\n", mi_s);
    fprintf(fipt, "# mi_s = %f\n", mi_s);
    fprintf(fout, "# cut_t = %f\n", max_t);
    fprintf(fout, "# displ = %.15f\n", displ);
    fprintf(fout, "# sali_c = %.15f\n", sali_c);
    fprintf(fout, "# Ni = %d\n", grid.gWidth);
    fprintf(fout, "# Nj = %d\n\n", grid.gHeight);
    fprintf(fipt, "# [i]  [j]    [x0]   [y0]   [px0]   [py0]   [H]\n");
    fprintf(fout, "# [SALI]    [i]  [j]    [x0]   [y0]   [px0]   [py0]   [H]\n");
    
    /* Set threads to use. 
     * if (nThreads==-1): use all available (dynamical) */    
    if (nThreads != -1 && 0 < nThreads && 10 >= nThreads)
    {
        omp_set_dynamic(0);
        omp_set_num_threads(nThreads);
    }

    /** Run **/
    int threads;
    int thread_id;
    int grid_pt = 0;
    int Nbar = 20;                     /* Progress bar size  */
    #pragma omp parallel private(thread_id, threads) shared(a_c, a_s, a_t, grid, grid_pt)
    {
    #pragma omp for schedule(dynamic) collapse(2)
    for (unsigned int i = 0; i < grid.gWidth; i++)
    {
        for (unsigned int j = 0; j < grid.gHeight; j++)
        {
            /* Progress print */
            #pragma omp critical
            {
                int grid_size = grid.gWidth * grid.gHeight;
                int color_ndx = 31;
                grid_pt += 1;
                threads   = omp_get_num_threads();
                thread_id = omp_get_thread_num();
			    cout << "\rSALI Grid point (" << grid.gWidth << " x " 
                     << grid.gHeight << "):   [p = " << grid_pt << " / " 
			    	 << grid_size << "]  ";
                if (grid_pt == grid_size) { color_ndx = 32; }
                print_prog_bar(Nbar, (double) grid_pt / (double) grid_size, color_ndx);
                cout << "(Thread " << thread_id + 1 << " / " << threads 
                     << ")  " << setfill(' ') << flush;
            }

            /* Set grid point */
            double x0  = x_min + ((double) i / (double) (grid.gWidth-1)) * fabs(x_max - x_min);
            double y0  = mi_s;
            double px0 = px_min + ((double) j / (double) (grid.gHeight-1)) * fabs(px_max - px_min);
            
            /* Set displacements */
            double dx0_1  = 0.0;
            double dy0_1  = displ;
            double dpx0_1 = 0.0;
            double dpy0_1 = 0.0;
            
            double dx0_2  = 0.0;
            double dy0_2  = 0.0;
            double dpx0_2 = 0.0;
            double dpy0_2 = displ;

            /* If it's a valid point -> run dynamics */
            if ((E - (pow(px0, 2) + V(x0, y0, U, a))) >= 0.0)
            {
                double py0 = sqrt(E - (pow(px0, 2) + V(x0, y0, U, a)));

                /** Dynamics settings **/
                double t  = 0.0;                   /* Dynamic time */
                double dt = 1e-4;                  /* Time step */
                double dt_max = 5e-3;              /* Max dt (maybe replace by dt_s) */
                double dt_s = 1.00;                /* Time step to evaluate SALI */
                double eps_abs = 1e-12;            /* Time step eps absolute control */
                double eps_rel = 1e-12;            /* Time step eps relative control */
                double varp[dimS];                 /* State variable array */
                double varn[] = {x0, y0, px0, py0, /* State variable array */
                                 dx0_1, dy0_1, dpx0_1, dpy0_1,
                                 dx0_2, dy0_2, dpx0_2, dpy0_2};
                double args[] = {a, U};

                /* Set GSL integration */
                const gsl_odeiv2_step_type *type = gsl_odeiv2_step_rkck;
                gsl_odeiv2_step *step = gsl_odeiv2_step_alloc(type, 3 * dimS);
                gsl_odeiv2_evolve *evolve  = gsl_odeiv2_evolve_alloc(3 * dimS);
                gsl_odeiv2_control *control = gsl_odeiv2_control_y_new(eps_abs, eps_rel);
                gsl_odeiv2_system sys = {system_funct_all, system_jacob_all, 3 * dimS, &args};

                /* Deviation vectors */
                vector<double> w1 = {varn[4], varn[5], varn[6], varn[7]};
                vector<double> w2 = {varn[8], varn[9], varn[10], varn[11]};
                double sali = SALI(w1, w2);
                vector<vector<double>> pss_pts = {{x0}, {y0}, {px0}, {py0}};
                
                int nPrint_s = 0;            /* Print SALI counter */
                bool empty = grid.check_empty_ndx(i, j);

                if (empty)
                {
                    double tmp_var[] = {varn[0], varn[1], varn[2], varn[3]};
                    grid.set_grid_pt_ndx(i, j, arr_to_vec(tmp_var, dimS));

                    fprintf(fipt, "%d  %d    %f   %f   %f   %f     %f\n", 
                            i, ((grid.gHeight-1) - j), varn[0], varn[1], varn[2], varn[3],
                                               Ham(varn[0], varn[1], varn[2], varn[3], U, a));

                    while (t < max_t)
                    {
                        /* Update previous step */
                        for (int k = 0; k < (int) dimS; k++) { varp[k] = varn[k]; }

                        /* Dynamical step */
                        int status = ode_step_cntrl(evolve, control, step, sys, 
                                                    t, dt, (t+dt_max), varn, 
                                                    mod_inf, mod_top, mod_inf, mod_top);
                        assert(status  == GSL_SUCCESS);
                        
                        /*   Evaluate SALI   */
                        /* Normalize vectors */
                        if (t > nPrint_s * dt_s)
                        {
                            double var1[] = {varn[4], varn[5], varn[6], varn[7]};
                            double var2[] = {varn[8], varn[9], varn[10], varn[11]};
                            w1 = arr_to_vec(var1, dimS);
                            w2 = arr_to_vec(var2, dimS);

                            sali = SALI(w1, w2);
                            nPrint_s += 1;

                            /* Pass normalized [w_i] to [var_i] */
                            for (int k = 0; k < (int) dimS; k++)
                            {
                                varn[k +     dimS] = w1[k];
                                varn[k + 2 * dimS] = w2[k];
                            }
                        }

                        /* Save point in PSS */
                        if ((varn[1] - mi_s) * (varp[1] - mi_s) < 0.0 && 
                             varp[3] > 0.0 && fabs(varp[1] - varn[1]) < PI)
                        {
                            double var_pss[dimS];
                            for (int k = 0; k < (int) dimS; k++) { var_pss[k] = varp[k]; }

                            int pss_status = pss_henon_step(dimS, mi_s, var_pss, args);
                            assert(0 == pss_status);
                            
                            /** Modulation **/
                            if (var_pss[0] >= mod_top) 
                            { 
                                var_pss[0] = var_pss[0] - (mod_top - mod_inf); 
                            } 
                            else if (var_pss[0] <= mod_inf) 
                            { 
                                var_pss[0] = var_pss[0] + (mod_top - mod_inf); 
                            }

                            /* Save point to avoid over calculation */
                            if (var_pss[0] >= grid.sX_min && var_pss[0] <= grid.sX_max &&
                                var_pss[2] >= grid.sY_min && var_pss[2] <= grid.sY_max)
                            {
                                for (int k = 0; k < (int) dimS; k++)
                                {
                                    pss_pts[k].push_back(var_pss[k]);
                                }
                            }
                        }

                        if (sali < sali_c) { break; }                        
                    }

                    /** Increment area **/
                    if (sali > sali_c)
                    {
                        a_s += cell_a;
                        a_t += cell_a;
                    }
                    else if (sali < sali_c)
                    {
                        a_c += cell_a;
                        a_t += cell_a;
                    }
                    
                    /** Print output **/
                    fprintf(fout, "%.15f    %d   %d     %f   %f   %f   %f     %f\n",
								  sali, i, ((grid.gHeight-1) - j), x0, y0, px0, py0, 
								  Ham(x0, y0, px0, py0, U, a));

                    /** Take extra PSS points **/ 
                    for (unsigned int p = 0; p < pss_pts[0].size(); p++)
                    {
                        vector<double> v = {pss_pts[0][p], pss_pts[1][p], 
                                            pss_pts[2][p], pss_pts[3][p]};

                        bool free = grid.check_empty_flt(v[0], v[2]);
                        if (free)
                        {
                            grid.set_grid_pt_flt(v[0], v[2], v);
                            if (sali > sali_c)
                            {
                                a_s += cell_a;
                                a_t += cell_a;
                            }
                            else if (sali < sali_c)
                            {
                                a_c += cell_a;
                                a_t += cell_a;
                            }

                            vector<int> ndx = grid.get_grid_ndx(v[0], v[2]);

                            fprintf(fout, "%.15f    %d   %d     %f   %f   %f   %f     %f\n",
                                    sali, ndx[0], ndx[1], v[0], v[1], v[2], v[3], 
                                                      Ham(v[0], v[1], v[2], v[3], U, a));
                        }
                    }
                }
                
                /* Free GSL */
                gsl_odeiv2_step_free(step);
                gsl_odeiv2_evolve_free(evolve);
                gsl_odeiv2_control_free(control);
            }
            else
            {
                fprintf(fout, "%.0f    %d   %d     %f   %f   %f   %f     %f\n",
                        -1.0, i, ((grid.gHeight-1) - j), x0, y0, px0, -1.0, -1.0);
            }
        }
    }
    }

    /** Append areas to output **/
    char fl_area_path[] = "area.dat";
    FILE *farea = fopen(fl_area_path, "a");

    if (farea == NULL)
    {
        printf("\nUnable to open '%s' file.\n", fl_area_path);
        printf("Please check whether file exists and you have write privilege.\n");
        exit(EXIT_FAILURE);
    }

    fprintf(farea, "%f   %f   %f   %f   %f   %f\n", E, a_s, a_c, a_t, a_s / a_t, a_c / a_t);

    /** Close files **/
    fclose(fout);
    fclose(fipt);
    fclose(farea);

    auto end = chrono::steady_clock::now();
    cout << "\nElapsed time: " << chrono::duration_cast<chrono::hours>(end - start).count() << "h  "
                             << chrono::duration_cast<chrono::minutes>(end - start).count() % 60 << "min  " 
                             << chrono::duration_cast<chrono::seconds>(end - start).count() % 60 << "sec \n" << endl;

    return 0;
}

void print_prog_bar(int bar_width, double progress, int ASCII_id)
{
    std::cout << "[";
    int pos = (int) (bar_width * progress);
    
    for (int i = 0; i < bar_width; i++)
    {
        if (i <= pos)
        {
            std::cout << "\033[" << ASCII_id << "m";
            std::cout << "=";
            std::cout << "\033[m";
        }
        else
        {
            std::cout << " ";
        }
    }

    std::cout << "] " << int(progress * 100) << " %  ";
}